"""
Data structures module
======================

Data types (e.g. Rows, Records) for ETL.
"""

from datetime import datetime
import uuid

MAX_LENGTH_POWER_OF_TWO = 16
INTEGER_LENGTHS = set(2 ** i for i in range(MAX_LENGTH_POWER_OF_TWO))
INTEGER_LENGTHS = INTEGER_LENGTHS | (set(i for i in range(32)))
VARCHAR_LENGTHS = set(2 ** i for i in range(MAX_LENGTH_POWER_OF_TWO))
VARCHAR_LENGTHS = VARCHAR_LENGTHS | (set(i for i in range(32)))


class IncompatibleTypesException(Exception):
    pass


class DataSourceTypeSystem:
    """
    Information about mapping one type system onto another contained in
    the children of this class.
    """

    @staticmethod
    def convert(obj):
        """
        Override this method if something more complicated is necessary.
        """
        obj = convert_to_type_system(obj, MySQLTypeSystem)
        return obj

    @staticmethod
    def type_mapping(*args, **kwargs):
        raise NotImplemented(
            "Class does not have a ``type_mapping`` function."
        )


def convert_to_type_system(obj, cls):
    members_of_type_system = [
        i
        for i in globals().values()
        if hasattr(i, "__bases__") and cls in all_bases(i)
    ]
    max_length = getattr(obj.original_type, "max_length", None)
    matching_max_length = [
        i
        for i in members_of_type_system
        if getattr(i, "max_length", None) == max_length
    ]
    matching_intermediate_data_type = [
        i for i in matching_max_length if i.intermediate_type is obj.__class__
    ]
    if len(matching_intermediate_data_type) == 0:
        raise Exception("No matching intermediate data type.")
    elif len(matching_intermediate_data_type) > 1:
        raise Exception("More than one matching intermediate data type")
    else:
        pass  # No other cases?
    data_type = matching_intermediate_data_type[0]
    converted_obj = data_type(obj.value)
    if hasattr(obj, "max_length"):
        converted_obj.max_length = obj.max_length
    converted_obj.original_type = obj.original_type
    return converted_obj


class PythonTypeSystem(DataSourceTypeSystem):
    pass


class PrimitiveTypeSystem(DataSourceTypeSystem):
    pass


class MySQLTypeSystem(DataSourceTypeSystem):
    """
    Each ``TypeSystem`` gets a ``type_mapping`` static method that takes a
    string and returns the class in the type system named by that string.
    For example, ``int(8)`` in a MySQL schema should return the
    ``MYSQL_INTEGER8`` class.
    """

    @staticmethod
    def type_mapping(string):
        """
        Parses the schema strings from MySQL and returns the appropriate class.
        """
        string = string.lower()
        if string.startswith("int") and "(" in string:
            max_length = string[4:-1]
            cls = globals()["MYSQL_INTEGER" + max_length]
        elif string.startswith("int"):
            cls = INTEGER_BASE
        elif string.startswith("varchar") and "(" in string:
            max_length = string[8:-1]
            cls = globals()["MYSQL_VARCHAR" + max_length]
        elif string.startswith("varchar"):
            cls = VARCHAR_BASE
        elif string == "date":
            cls = MYSQL_DATE
        else:
            cls = MYSQL_VARCHAR128
            # raise Exception('Unrecognized MySQL type: {type_string}'.format(
            #    type_string=string))
        return cls


class IntermediateTypeSystem(DataSourceTypeSystem):
    """
    Never instantiate this by hand.
    """

    pass


def primitive_to_intermediate_type(thing, name=None):
    if isinstance(thing, (int,)):
        cast_type = INTEGER
    elif isinstance(thing, (float,)):
        cast_type = FLOAT
    elif isinstance(thing, (str,)):
        cast_type = STRING
    else:
        raise Exception("Unknown type")
    return cast_type(thing, name=name, original_type=PrimitiveTypeSystem)


class DataType:
    """
    Each ``DataType`` gets a ``python_cast_function``, which is a function.
    """

    python_cast_function = None
    intermediate_type = None

    def __init__(self, value, original_type=None, name=None):
        self.value = value
        self.name = name or uuid.uuid4().hex
        self.original_type = original_type

    @classmethod
    def __repr__(cls):
        return "hi"

    def __repr__(self):
        return ":".join([str(self.value), self.__class__.__name__])

    def to_intermediate_type(self):
        """
        Convert the ``DataType`` to an ``IntermediateDataType`` using its
        class's ``intermediate_type`` attribute.
        """
        if self.__class__.intermediate_type is None:
            raise Exception("No ``intermediate_type`` cast defined.")
        return self.__class__.intermediate_type(
            self.value, original_type=self.__class__, name=self.name
        )

    def to_python(self):
        if self.__class__.python_cast_function is None:
            raise Exception("No method for casting to Python primitive.")
        else:
            return self.__class__.python_cast_function(self.value)

    @property
    def type_system(self):
        """
        Just for convenience to make the type system an attribute.
        """
        return get_type_system(self)


class STRING(DataType, IntermediateTypeSystem):
    python_cast_function = str


class INTEGER(DataType, IntermediateTypeSystem):
    python_cast_function = int


class DATETIME(DataType, IntermediateTypeSystem):
    python_cast_function = lambda x: x


class FLOAT(DataType, IntermediateTypeSystem):
    python_cast_function = float


class BOOL(DataType, IntermediateTypeSystem):
    python_cast_function = bool


# MYSQL TYPES
#
# Each ``DataType`` has a ``python_cast_function`` and an ``intermediate_type``
# attribute. The ``intermediate_type`` is the class in the
# ``IntermediateTypeSystem`` to which this ``DataType`` would be cast.


class MYSQL_VARCHAR_BASE(DataType, MySQLTypeSystem):
    python_cast_function = str
    intermediate_type = STRING


class MYSQL_ENUM(DataType, MySQLTypeSystem):
    python_cast_function = str  # Placeholder
    intermediate_type = STRING  # Needs to be changed


class MYSQL_DATE(DataType, MySQLTypeSystem):
    python_cast_function = lambda x: datetime.datetime.strptime(x, "%Y-%m-%d")
    intermediate_type = DATETIME


class MYSQL_BOOL(DataType, MySQLTypeSystem):
    python_cast_function = bool
    intermediate_type = BOOL


class MYSQL_VARCHAR(type):
    def __new__(cls, max_length):
        x = super().__new__(
            cls,
            "MYSQL_VARCHAR{max_length}".format(max_length=str(max_length)),
            (MYSQL_VARCHAR_BASE,),
            {"max_length": max_length},
        )
        return x


class MYSQL_INTEGER_BASE(DataType, MySQLTypeSystem):
    python_cast_function = int
    intermediate_type = INTEGER


class MYSQL_INTEGER(type):
    def __new__(cls, max_length):
        x = super().__new__(
            cls,
            "MYSQL_INTEGER{max_length}".format(max_length=str(max_length)),
            (MYSQL_INTEGER_BASE,),
            {"max_length": max_length},
        )
        return x


def make_types():
    types_dict = {}
    for varchar_length in VARCHAR_LENGTHS:
        types_dict["MYSQL_VARCHAR" + str(varchar_length)] = MYSQL_VARCHAR(
            varchar_length
        )

    for integer_length in INTEGER_LENGTHS:
        types_dict["MYSQL_INTEGER" + str(integer_length)] = MYSQL_INTEGER(
            integer_length
        )
    return types_dict


globals().update(make_types())


def mysql_type(string):
    """
    Parses the schema strings from MySQL and returns the appropriate class.
    """
    string = string.lower()
    if string.startswith("int") and "(" in string:
        max_length = string[4:-1]
        try:
            cls = globals()["MYSQL_INTEGER" + max_length]
        except:
            import pdb

            pdb.set_trace()
    elif string.startswith("int"):
        cls = INTEGER_BASE
    elif string.startswith("varchar") and "(" in string:
        max_length = string[8:-1]
        cls = globals()["MYSQL_VARCHAR" + max_length]
    elif string.startswith("varchar"):
        cls = VARCHAR_BASE
    elif string == "date":
        cls = MYSQL_DATE
    else:
        cls = MYSQL_VARCHAR128  # No
        # raise Exception('Unrecognized MySQL type: {type_string}'.format(
        #    type_string=string))
    return cls


class Row:
    """
    A collection of ``DataType`` objects (typed values). They are dictionaries
    mapping the names of the values to the ``DataType`` objects.
    """

    def __init__(self, *records, type_system=None):
        """
        Constructor for ``Row``.

        Args:
            records: A list of ``DataType`` objects
        """

        self.records = {record.name: record for record in records}
        self.type_system = type_system

    def is_empty(self):
        return len(self.records) == 0

    def __getattr__(self, attr):
        """
        Overrides the usual ``__getattr__`` method. The purpose of this
        is to make each ``Record`` in the ``Row`` accessible as an attribute.
        """

        if attr not in self.__dict__ and attr in self.records:
            return self.records[attr]
        return super(Row, self).__getattr__(attr)

    def __repr__(self):
        return ", ".join([str(record) for record in self.records.values()])

    @staticmethod
    def from_dict(row_dictionary, **kwargs):
        """
        Creates a ``Row`` object form a dictionary mapping names to values.
        """

        import pdb

        pdb.set_trace()
        return Row(*record_list)

    def concat(self, other, fail_on_duplicate=True):
        if len(set(self.keys()) & set(other.keys())) > 0 and fail_on_duplicate:
            raise Exception(
                "Overlapping records during concatenation of `Row`."
            )
        self.records.update(other.records)
        return self

    def keys(self):
        """
        For implementing the mapping protocol.
        """
        return self.records.keys()

    def __getitem__(self, key):
        """
        Cast to a dictionary and get back Python types.
        """
        obj = getattr(self, key)
        return obj.to_python()

    def __del__(self, key):
        del self.records[key]

    def __iadd__(self, other):
        if self.type_system is not other.type_system:
            raise IncompatibleTypesException(
                """Tried to concatenate `Row` objects with incompatible """
                """type systems {type_1} and {type_2}.""".format(
                    type_1=self.type_system.__name__,
                    type_2=other.type_system.__name__,
                )
            )
        concat(self, other, fail_on_duplicate=False)


def all_bases(obj):
    """
    Return all the class to which ``obj`` belongs.
    """

    def _inner(thing, bases=None):
        bases = bases or set()
        if not hasattr(thing, "__bases__"):
            thing = thing.__class__
        for i in thing.__bases__ or []:
            bases.add(i)
            bases = bases | _inner(i, bases=bases)
        return bases

    return set(_inner(obj))


def get_type_system(obj):
    bases = all_bases(obj)
    type_system_list = [
        i for i in bases if DataSourceTypeSystem in i.__bases__
    ]
    if len(type_system_list) > 1:
        raise Exception("Belongs to more than one type system?")
    elif len(type_system_list) == 0:
        raise Exception("Belongs to no type system?")
    else:
        return type_system_list[0]


if __name__ == "__main__":
    bar = MYSQL_VARCHAR16("bar", name="bar")
    bar = MYSQL_INTEGER8(12)
    r = Row(bar, bar, type_system=MySQLTypeSystem)
    t = Row(type_system=PythonTypeSystem)
    intermediate = bar.to_intermediate_type()
    converted_bar = MySQLTypeSystem.convert(intermediate)



import sys
import logging

logging.basicConfig(level=logging.INFO)


class NoisyImportFinder:

    PATH_TRIGGER = "__metalpipe__"

    def __init__(self, path_entry):
        self.path_entry = path_entry
        print("Checking {}:".format(path_entry), end=" ")
        if path_entry != self.PATH_TRIGGER:
            print("wrong finder")
            raise ImportError()
        else:
            print("works")
        return

    def find_module(self, fullname, path=None):
        logging.info(fullname)
        if fullname == self.PATH_TRIGGER:
            logging.info("Loading derived class")
            import pdb

            pdb.set_trace()

        return None


if __name__ == "__main__":
    sys.path_hooks.append(NoisyImportFinder)

    for hook in sys.path_hooks:
        print("Path hook: {}".format(hook))

    sys.path.insert(0, NoisyImportFinder.PATH_TRIGGER)

    try:
        print("importing target_module")
        import __metalpipe__.bar
    except Exception as e:
        print("Import failed:", e)



"""
MetalPipeMessage module
========================

The ``MetalPipeMesaage`` encapsulates the content of each piece of data,
along with some useful metadata.
"""

import time
import uuid
import hashlib


class MetalPipeMessage(object):
    """
    A class that contains the message payloads that are queued for
    each ``MetalPipeProcessor``. It holds the messages and lots
    of metadata used for logging, monitoring, etc.
    """

    def __init__(self, message_content):
        if not isinstance(message_content, (dict,)):
            raise Exception(
                "Message content must be a dictionary or "
                "`output_keypath` must be specified."
            )
        self.message_content = message_content
        self.message_hash = hashlib.md5(
            bytes(str(message_content), encoding="utf8")
        )
        self.history = []
        self.time_created = time.time()
        self.time_queued = None
        self.time_processed = None
        self.uuid = uuid.uuid4()
        self.accumulator = {}

    def __repr__(self):
        s = ": ".join(["MetalPipeMessage", self.uuid.hex])
        return s



"""
MetalPipeQueue module
=====================

These are queues that form the directed edges between nodes.
"""

import queue
import uuid
import logging
import time
from metalpipe.message.message import MetalPipeMessage
from metalpipe.message.batch import BatchStart, BatchEnd
from metalpipe import node


QUEUE_TIME_WINDOW = 100


class MetalPipeQueue:
    """
    """

    def __init__(self, max_queue_size, name=None):
        self.queue = queue.Queue(max_queue_size)
        self.max_queue_size = max_queue_size
        self.name = name or uuid.uuid4().hex
        self.source_node = None
        self.target_node = None
        self.queue_times = []  # Time messages spend in queue

    def size(self):
        return self.queue.qsize()

    def approximately_full(self, error=0.95):
        return self.size() >= (self.max_queue_size - 1)

    @property
    def empty(self):
        return self.queue.empty()

    def get(self):
        try:
            message = self.queue.get(block=False)
            self.queue_times.append(time.time() - message.time_queued)
            self.queue_times = self.queue_times[-1 * QUEUE_TIME_WINDOW :]
            logging.debug(
                "QUEUE TIMES: {queue_times}".format(
                    queue_times=str(self.queue_times)
                )
            )
            logging.debug(
                "QUEUE SIZE: {queue_size}".format(queue_size=str(self.size()))
            )
            logging.debug(
                "QUEUE FULL: {queue_full}".format(
                    queue_full=str(self.approximately_full())
                )
            )
        except queue.Empty:
            message = None
        logging.debug("Retrieved message: " + str(message))
        return message

    def put(self, message, *args, previous_message=None, **kwargs):
        """
        Places a message on the output queues. If the message is ``None``,
        then the queue is skipped.

        Messages are ``MetalPipeMessage`` objects; the payload of the
        message is message.message_content.
        """
        if isinstance(message, (node.NothingToSeeHere,)):
            return
        elif previous_message is not None:
            previous_message = previous_message.message_content
        else:
            previous_message = {}
        if self.source_node.output_key is not None:
            message = {self.source_node.output_key: message}

        # Check if we need to retain the previous message in the keys of
        # this message, assuming we have dictionaries, etc.
        if self.source_node.retain_input:
            for key, value in previous_message.items():
                if key not in message:
                    message[key] = value
                elif (
                    key in message
                    and value is not None
                    and self.source_node.prefer_existing_value
                ):
                    message[key] = value
                else:
                    pass

        message_obj = MetalPipeMessage(message)
        message_obj.time_queued = time.time()
        self.queue.put(message_obj)



"""
A decorator that allows you to specify a set of keyword arguments that
are required by the function. If any are missing, an exception is
raised which lists the missing arguments.

::

    @required_arguments(bar='bar', baz='qux')
    def my_function(x, y, bar=None, baz=None):
        ...something...


Invoking the function like so:

::

    my_function(1, 2, bar='barbar')


will raise:

::

    MissingRequiredArgument: Missing required argument(s): bar

"""


class MissingRequiredArgument(Exception):
    pass


class required_arguments:
    def __init__(self, *kwarg_list):
        self.kwarg_list = kwarg_list

    def __call__(self, f):
        def inner_function(*args, **kwargs):
            missing_kwargs = []
            for kwarg in self.kwarg_list:
                if kwarg not in kwargs:
                    missing_kwargs.append(kwarg)
            if len(missing_kwargs) > 0:
                raise MissingRequiredArgument(
                    "Missing required argument(s): "
                    "{kwarg_list}".format(kwarg_list=", ".join(missing_kwargs))
                )
            out = f(*args, **kwargs)
            return out

        return inner_function



# Borked!
"""
Experiment using `bowerbird` to set up a flexible watchdog job that
looks for new files to appear and sends their names downstream. Goal is
to have a single, configurable job that can watch files systems, sftp,
S3, GCP, etc.

"""

import time
import logging
import re

# from bowerbird.filesystem import LocalFileSystem



"""
Helper module
*************

Misc. helper functions for other classes.
"""

import copy
import types
import time
import logging
import importlib
import datetime
import pickle
import hashlib
import uuid
import pickle
import base64


UNIX_EPOCH = datetime.datetime(month=1, year=1970, day=1)


def package(thing):
    return base64.b64encode(pickle.dumps(thing))


def unpackage(thing):
    return pickle.loads(base64.b64decode(thing))


def list_to_dict(some_list, list_of_keys):
    if len(some_list) != len(list_of_keys):
        raise Exception("Length of list elements and key list must be equal.")
    out = {list_of_keys[index]: item for index, item in enumerate(some_list)}
    return out


def load_function(function_name):
    components = function_name.split("__")
    module = ".".join(components[:-1])
    function_name = components[-1]
    module = importlib.import_module(module)
    function = getattr(module, function_name)
    return function


def timestamp_to_redshift(timestamp):
    if isinstance(timestamp, (str,)):
        return timestamp
    return timestamp.strftime("%b %d,%Y  %H:%M:%S")


def string_to_redshift(timestamp):
    """
    TODO: Consider removing this function.
    """
    if timestamp is None or (
        isinstance(timestamp, (str,)) and timestamp == ""
    ):
        return timestamp
    return datetime.datetime.strptime(timestamp, "%b %d,%Y  %H:%M:%S")


def string_to_datetime(timestamp):
    if timestamp is None or (
        isinstance(timestamp, (str,)) and timestamp == ""
    ):
        return timestamp
    return datetime.datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S")


def milliseconds_epoch_to_datetime(milliseconds_epoch):
    if isinstance(milliseconds_epoch, (datetime.datetime,)):
        return milliseconds_epoch
    try:
        milliseconds_epoch = int(milliseconds_epoch)
    except:
        logging.info(
            "milliseconds exception: "
            + str(type(milliseconds_epoch))
            + " "
            + str(milliseconds_epoch)
        )
        return milliseconds_epoch
    logging.debug("milliseconds_epoch_to_datetime: " + str(milliseconds_epoch))
    out = UNIX_EPOCH + datetime.timedelta(
        seconds=(int(milliseconds_epoch) / 1000)
    )
    logging.debug("milliseconds_epoch_to_datetime output: " + str(out))
    return out


def seconds_epoch_to_datetime(seconds_epoch):
    return UNIX_EPOCH + datetime.timedelta(seconds=(seconds))


def to_bool(thing):
    if isinstance(thing, (str,)):
        return len(thing) > 0 and thing[0].lower() in ["t", "y"]
    elif isinstance(thing, (int, float)):
        return thing > 0
    elif isinstance(thing, (bool,)):
        return thing
    else:
        raise Exception(
            "Do not know how to convert {thing} to bool".format(
                thing=str(thing)
            )
        )


def get_value(
    dictionary,
    path,
    delimiter=".",
    use_default_value=False,
    default_value=None,
):

    # dictionary = copy.deepcopy(dictionary)
    if isinstance(path, (str,)):
        path = path.split(delimiter)
    elif isinstance(path, (list, tuple)):
        pass
    else:
        raise Exception("what?")
    if len(path) == 0:
        dictionary = dictionary
    else:
        for step in path:
            if dictionary is None or isinstance(dictionary, (dict,)):
                dictionary = (dictionary or {}).get(step, default_value)
            else:
                dictionary = dictionary[step.index]
    return dictionary


def hi(*args, **kwargs):
    return "hi"


def pdb(*args, **kwargs):
    import pdb

    pdb.set_trace()


def convert_date_format(date_string, source_format=None, target_format=None):
    return datetime.datetime.strptime(date_string, source_format).strftime(
        target_format
    )


def engaging_networks_date(result, **kwargs):
    """
    This is a very redundant function, but it's here mostly to test the
    functionality of the ``post_process_function``.
    """
    out = datetime.datetime.strptime(result[0]["date_modified"], "%Y-%m-%d")
    out = out.strftime("%Y%m%d")
    return out


def set_value(dictionary, path, value):
    for step in path[:-1]:
        if not isinstance(step, (ListIndex,)):
            if step not in dictionary:
                raise Exception("Path not found.")
            dictionary = dictionary[step]
        else:
            dictionary = dictionary[step.index]
    dictionary[
        path[-1] if not isinstance(path[-1], (ListIndex,)) else path[-1].index
    ] = value


def iterate_leaves(dictionary, keypath=None):
    keypath = keypath or []
    for key, value in dictionary.items():
        if not isinstance(value, (dict,)):
            yield keypath + [key], value
        else:
            for i in iterate_leaves(value, keypath=keypath + [key]):
                yield i


def remap_dictionary(
    source_dictionary,
    target_dictionary,
    use_default_value=False,
    default_value=None,
):
    target_dictionary = copy.deepcopy(target_dictionary)
    for path, value in iterate_leaves(target_dictionary):
        set_value(
            target_dictionary,
            path,
            get_value(
                source_dictionary,
                value,
                use_default_value=use_default_value,
                default_value=default_value,
            ),
        )
    return target_dictionary


def now_milliseconds():
    return str(int(time.time() * 1000))


def two_weeks_ago():
    return (
        datetime.datetime(year=2019, month=1, day=1) - UNIX_EPOCH
    ).total_seconds() * 1000
    # return str(int(time.time() * 1000 - (30 * (24 * 60 * 60 * 1000))))


def now_redshift():
    return datetime.datetime.now().strftime("%b %d,%Y  %H:%M:%S")


def now_datetime():
    return datetime.datetime.now()


def two_weeks_ago_datetime():
    return datetime.datetime.now() - datetime.timedelta(days=14)


def january_1_2019():
    return datetime.datetime(year=2016, month=1, day=1)


def datetime_to_redshift(datetime_obj):
    return datetime_obj.strftime("%b %d,%Y  %H:%M:%S")


def datetime_to_milliseconds(datetime_obj):
    return int((datetime_obj - UNIX_EPOCH).total_seconds() * 1000)


class SafeMap(dict):
    def __missing__(self, key):
        return "{" + str(key) + "}"


class ListIndex:
    def __init__(self, index):
        self.index = index

    def __repr__(self):
        return "ListIndex({index})".format(index=str(self.index))

    def __eq__(self, other):
        return isinstance(other, (ListIndex,)) and self.index == other.index

    def __hash__(self):
        return self.index


def all_paths(thing, path=None, starting_path=None):
    path = path or tuple()
    starting_path = starting_path or []
    thing = get_value(thing, starting_path)

    if isinstance(thing, (dict,)):
        for key, value in thing.items():
            yield path + (key,)
            for i in all_paths(value, path=path + (key,)):
                yield i

    elif isinstance(thing, (list,)):
        for index, item in enumerate(thing):
            yield path + (ListIndex(index),)
            for i in all_paths(item, path=path + (ListIndex(index),)):
                yield i
    else:
        yield path


def last_element_list_index(path):
    return isinstance(path[-1], (ListIndex,))


def remove_list_indexes(path):
    return tuple(step for step in path if not isinstance(step, (ListIndex,)))


def matching_tail_paths(target_path, structure, starting_path=None):
    target_path = tuple(target_path)
    if starting_path is not None:
        starting_path = tuple(starting_path)
    seen = set()
    for path in all_paths(structure, starting_path=starting_path):
        if path in seen:
            continue
        elif len(path) == 0:
            continue
        elif last_element_list_index(path):
            continue
        else:
            seen.add(path)

        temp_list = remove_list_indexes(path)

        if len(temp_list) < len(target_path):
            continue
        tail_of_path = temp_list[-1 * len(target_path) :]
        if lists_equal(tail_of_path, target_path):
            logging.debug("tail path: " + str(target_path) + " " + str(path))
            yield (starting_path or tuple([])) + path


def lists_equal(l1, l2):
    return l1 == l2


def nevermind(thing):
    return thing


def replace_by_path(
    dictionary,
    target_path,
    target_value=None,
    function=None,
    function_args=None,
    function_kwargs=None,
    starting_path=None,
):

    function = function or nevermind
    function_args = function_args or tuple([])
    function_kwargs = function_kwargs or {}
    # dictionary_clone = copy.deepcopy(dictionary)
    dictionary_clone = dictionary
    for path in matching_tail_paths(
        target_path, dictionary_clone, starting_path=starting_path
    ):
        current_value = get_value(dictionary, path)
        one_target_value = target_value or function(
            current_value, *function_args, **function_kwargs
        )
        set_value(dictionary, path, one_target_value)


def aggregate_values(dictionary, target_path, values=False):
    aggregated_values = []
    for path in matching_tail_paths(target_path, dictionary):
        current_value = get_value(dictionary, path)
        try:
            aggregated_values.append(
                list(current_value.values()) if values else current_value
            )
        except AttributeError:
            logging.debug(
                "Attribute error in aggregate values. No list returned."
            )

    logging.debug("aggregated_values: " + str(aggregated_values))
    try:
        out = aggregated_values if not values else aggregated_values[0]
    except IndexError:
        out = []
    return out


class UberDict(dict):
    def rget(self, value):
        for i in meets_condition(self, lambda x: isinstance(x, (dict,))):
            for _key, _value in i.items():
                if _value == value:
                    yield _key


def iterate(thing, path=None, seen=None, start_path=None):
    """
    TODO: allow specification of `start_path` to speed this up.
    """
    hashed_thing = hash(str(thing))
    path = path or []
    start_path = start_path or []
    seen = seen or set([])
    if hashed_thing not in seen:
        if isinstance(thing, (list, tuple, dict)):
            yield thing
        seen.add(hashed_thing)
        if isinstance(thing, (dict,)):
            for key, value in thing.items():
                for item in iterate(value, path=path + [key], seen=seen):
                    yield item
        elif isinstance(thing, (list, tuple)):
            for index, item in enumerate(thing):
                for thingie in iterate(
                    item, path=path + [ListIndex(index)], seen=seen
                ):
                    yield thingie
        else:
            yield thing


def meets_condition(thing, func):
    for item in iterate(thing):
        if func(item):
            yield item


if __name__ == "__main__":
    import json

    d = json.load(open("./sample.json", "r"))
    counter = 0
    while counter < 100:
        replace_by_path(d, ["vid"], target_value="hithere")
        counter += 1
    import random

    logging.basicConfig(level=logging.debug)
    replace_by_path(
        d, ["vid"], function=lambda x: str(x) + str(random.random())
    )



"""
Trigger module
==============

A simple class containing no data, which is intended merely as a trigger,
signaling that the downstream node should do something.
"""

import time
import random
import datetime
import logging
import re
import hashlib
import schedule


def hello_world():
    print("hello, world!")


class Trigger:
    def __init__(self, previous_trigger_time=None, trigger_name=None):
        self.previous_trigger_time = None
        self.trigger_name = (
            trigger_name
            or hashlib.md5(bytes(str(random.random()), "ascii")).hexdigest()
        )
        self.time_sent = time.time()  # In epochs
        logging.debug(
            "Sent trigger at {now}".format(now=str(datetime.datetime.now()))
        )



"""
Table Nodes Module
==================

Analyzing tables, converting types, etc. Not for interacting directly with
tabular databases
"""

import requests
import collections
import json
import time
import random
import logging
from metalpipe.node import MetalNode, NothingToSeeHere
from bloom_filter import BloomFilter
from metalpipe.utils.data_structures import make_types


class IntermediateDataType:
    def __init__(self):
        pass


class VarChar(IntermediateDataType):
    pass


PYTHON_PRIMITIVE_DATA_TYPES = [
    int,
    float,
    str,
    bytes,]


def get_type_information(thing):
    thing = str(thing)
    possible_types = []
    for data_type in PYTHON_PRIMITIVE_DATA_TYPES:
        try:
            data_type(thing)
            possible_types.append(data_type)
        except:
            pass
    return None


class RowStatCollector(MetalNode):

    def __init__(self, **kwargs):
        self.bloom_filter_dict = collections.defaultdict(BloomFilter)
        self.uniqueness_dict = collections.defaultdict(bool)
        self.num_unique_values_dict = collections.defaultdict(int)
        self.null_dict = collections.defaultdict(bool)
        self.first_row = True
        super(RowStatCollector, self).__init__(**kwargs)

    def process_item(self):
        if isinstance(self.__message__, (dict,)):
            rows = [self.__message__]
        elif isinstance(self.__message__, (list, tuple, set,)):
            rows = self.__message__
        else:
            raise Exception('This should not happen.')
        for row in rows:
            if self.first_row:
                for column, _ in row.items():
                    self.uniqueness_dict[column] = True  # Default unique to True
            for column, value in row.items():
                # Use bloom filter to check for unique values
                if value in self.bloom_filter_dict[column]:
                    self.uniqueness_dict[column] = False
                else:
                    self.num_unique_values_dict[column] += 1
                    self.uniqueness_dict[column].add(value)

                # Record whether a NULL occurs in the column
                if value is None:
                    self.null_dict[column] = False
        yield NothingToSeeHere()

    def cleanup():
        pass




"""
"""

import pickle as pickle
import base64


def encode(something):
    """
    We encode all messages as base64-encoded pickle objects in case
    later on, we want to persist them or send them to another system.
    This is extraneous for now.
    """
    return base64.b64encode(pickle.dumps(something))


def decode(something):
    """
    Decodes from base-64 pickled object.
    """
    return pickle.loads(base64.b64decode(something))



import hashlib
import redis

from metalpipe.utils.helpers import package, unpackage


class RedisFixturizer:
    def __init__(self, db=0, host="localhost", port=6379):
        self.db = db
        self.host = host
        self.port = port
        self.redis = redis.Redis(host=host, port=port, db=db)

    def get_num_source_messages(self, node):
        node_key = "__num_values_stored_{node_name}__".format(
            node_name=node.name
        )
        try:
            num_messages = self.redis.get(node_key)
        except:
            num_messages = -1
        if num_messages is None:
            num_messages = -1
        return num_messages

    def record_source_node(self, node, value):
        num_messages = int(self.get_num_source_messages(node))
        message_number = num_messages + 1
        key = "{node_name}_{message_number}".format(
            node_name=node.name, message_number=str(message_number)
        )
        self.redis.set(key, package(value))
        node_key = "__num_values_stored_{node_name}__".format(
            node_name=node.name
        )
        self.redis.set(node_key, message_number)

    def record_worker_node(self, node, input_value, output_value):
        input_hash = hashlib.md5(bytes(str(input_value), "utf8")).hexdigest()
        key = "{node_name}_{input_hash}".format(
            node_name=node.name, input_hash=input_hash
        )
        try:
            value = self.redis.get(key)
        except:
            value = None
        value = unpackage(value) if value is not None else []
        value.append(output_value)
        self.redis.set(key, package(value))


if __name__ == "__main__":

    from metalpipe.node import *
    from metalpipe.node_classes.network_nodes import *
    from metalpipe.node_classes.civis_nodes import *

    node_1 = ConstantEmitter(
        name="node_1",
        delay=0.01,
        max_loops=200,
        thing={"hi": "there"},
        fixturize=True,
    )
    node_2 = PrinterOfThings(fixturize=True, name="node_2")
    node_3 = PrinterOfThings(fixturize=True, name="node_3")
    node_1 > node_2 > node_3

    # record = MetalPipeRecorder(record_to_file="test_file.fs", pipeline=node_1)
    node_1.global_start(fixturize=True)
    node_1.wait_for_pipeline_finish()
    sys.exit()



"""
Network nodes module
====================

Classes that deal with sending and receiving data across the interwebs.
"""

import requests
import json
import time
import random
import logging
from requests.packages.urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter

from metalpipe.node import MetalNode
from metalpipe.utils.helpers import SafeMap

additional_data_test = bool


class PaginatedHttpGetRequest:
    """
    For handling requests in a semi-general way that require paging through
    lists of results and repeatedly making GET requests.
    """

    def __init__(
        self,
        endpoint_template=None,
        additional_data_key=None,
        pagination_key=None,
        pagination_get_request_key=None,
        protocol="http",
        retries=5,
        default_offset_value="",
        additional_data_test=bool,
        calling_node=None,
    ):
        """
        :ivar endpoint_template: (str) Template for endpoint URL, suitable
            for calling ``endpoint_template.format(**kwargs)``.
        :ivar additional_data_key: Key in JSON payload whose value
            indicates whether there are additional pages to request.
        :ivar pagination_key: Key in JSON payload where the current page
            (or next page) is indicated.
        :ivar pagination_get_request_key: Variable in URL GET request where
            we pass the offset for the next page.
        :ivar default_offset_value: Offset value for the first request.
            Usually this will be an empty string.
        :ivar additional_data_test: Function that is passed the value of
            ``additional_data_key``. It should return ``True`` if there are
            additional pages to request, ``False`` otherwise.
        """
        self.endpoint_template = endpoint_template
        self.additional_data_key = additional_data_key
        self.pagination_key = pagination_key
        self.pagination_get_request_key = pagination_get_request_key
        self.protocol = protocol
        self.retries = retries
        self.default_offset_value = default_offset_value
        self.additional_data_test = additional_data_test

    def get_with_retry(self, url, error_on_none=True, **kwargs):
        """
        Simple method for making requests from flaky endpoints.
        """

        error_counter = 0
        success = False
        hibernate = 1.0
        while error_counter < self.retries and not success:
            try:
                output = requests.get(url)
                if output is None:
                    logging.info(
                        "Request to {url} returned None".format(url=url)
                    )
                elif output.status_code >= 300:
                    logging.info(
                        "Request to {url} returned {code} status code".format(
                            url=url, code=str(output.status_code)
                        )
                    )
                else:
                    success = True
            except:
                error_counter += 1
                time.sleep(hibernate)
                hibernate *= 2
        if success:
            return output
        else:
            raise Exception("Failure for URL: {url}".format(url=url))

    def responses(self):
        """
        Generator. Yields each response until empty.
        """

        GET_ONLY = True

        offset_set = set()
        session = requests.Session()
        retries = Retry(
            total=self.retries,
            read=self.retries,
            connect=self.retries,
            backoff_factor=0.3,
            status_forcelist=[500, 502, 503, 504],
        )

        session.mount(
            "{protocol}://".format(protocol=self.protocol),
            HTTPAdapter(max_retries=retries),
        )

        get_request_parameters = {
            self.pagination_get_request_key: self.default_offset_value
        }
        endpoint_url = self.endpoint_template.format(**get_request_parameters)
        logging.info("paginator url: " + endpoint_url)
        successful = False
        retry_counter = 0
        sleep_time = 1.0
        while not successful and retry_counter <= self.retries:
            try:
                out = requests.get(endpoint_url)
                successful = True
            except:
                logging.info(
                    "sleeping randomly... retry: {retry}".format(
                        retry=str(retry_counter)
                    )
                )
                retry_counter += 1
                time.sleep(sleep_time + (random.random() * 2))

        # Check if successful
        if not successful:
            logging.info(
                "Unsuccessful request to {url}".format(url=endpoint_url)
            )
            raise Exception("Unsuccessful GET request")

        # out = out.json()
        out = json.loads(out.text)
        offset = out.get(self.pagination_key, None)
        offset_set.add(offset)
        page_counter = 0
        while self.additional_data_key in out and additional_data_test(
            out[self.additional_data_key]
        ):
            try:
                offset = out[self.pagination_key]
                offset_set.add(offset)
            except KeyError:
                logging.debug("No offset key. Assuming this is normal.")
                break
            get_request_parameters = {self.pagination_get_request_key: offset}
            endpoint_url = self.endpoint_template.format(
                **get_request_parameters
            )
            logging.debug("paginator url: " + endpoint_url)
            try:
                # response = session.get(endpoint_url)

                if GET_ONLY:
                    response = self.get_with_retry(endpoint_url)
                else:
                    response = session.get(endpoint_url)

                out = response.json()
            except:
                logging.warning(
                    "Error parsing. Assuming this is the "
                    "end of the responses."
                )
                break
            yield out


class HttpGetRequest(MetalNode):
    """
    Node class for making simple GET requests.
    """

    def __init__(
        self,
        endpoint_template=None,
        endpoint_dict=None,
        protocol="http",
        retries=5,
        json=True,
        **kwargs
    ):
        self.endpoint_template = endpoint_template
        self.endpoint_dict = endpoint_dict or {}
        self.protocol = protocol
        self.json = json
        self.retries = retries
        self.endpoint_dict.update(self.endpoint_dict)

        super(HttpGetRequest, self).__init__(**kwargs)

    def process_item(self):
        """
        The input to this function will be a dictionary-like object with
        parameters to be substituted into the endpoint string and a
        dictionary with keys and values to be passed in the GET request.

        Three use-cases:
        1. Endpoint and parameters set initially and never changed.
        2. Endpoint and parameters set once at runtime
        3. Endpoint and parameters set by upstream messages
        """

        # Hit the parameterized endpoint and yield back the results
        formatted_endpoint = self.endpoint_template.format_map(
            SafeMap(**(self.message or {}))
        )
        try:
            formatted_endpoint = formatted_endpoint.format_map(
                SafeMap(**(self.endpoint_dict or {}))
            )
        except Exception:
            logging.error("formatted endpoint: " + formatted_endpoint)
            raise Exception()
        self.log_info(
            "Http GET request: {endpoint}".format(endpoint=formatted_endpoint)
        )

        session = requests.Session()
        retries = Retry(
            total=self.retries,
            backoff_factor=0.1,
            status_forcelist=[500, 502, 503, 504],
        )
        session.mount(
            "{protocol}://".format(protocol=self.protocol),
            HTTPAdapter(max_retries=retries),
        )

        get_response = session.get(formatted_endpoint)
        try:
            output = get_response.json()
        except json.JSONDecodeError:
            output = get_response.text
        logging.debug(
            formatted_endpoint
            + " GET RESPONSE: "
            + str(output)
            + str(type(output))
        )
        yield output


class HttpGetRequestPaginator(MetalNode):
    """
    Node class for HTTP API requests that require paging through sets of
    results.

    This class handles making HTTP GET requests, determining whether there
    are additional results, and making additional calls if necessary. A typical
    case is to have an HTTP request something like this:

    .. code-block:: none

        http://www.someapi.com/endpoint_name?resultpage=0

    with a response like:

    .. code-block:: none

        {"data": "something", "additional_pages": true, "next_page": 1}

    The response contains some data, a flag ``additional_pages`` for determining
    whether there are additional results, and a parameter that gets passed to
    the next request for retrieving the right page of results (``next_page``).
    So the next GET request would be:

    .. code-block:: none

        http://www.someapi.com/endpoint_name?resultpage=1

    This process will repeat until ``additional_pages`` is false.

    In order to use this node class, you'll need to provide arguments that
    tell the node where to look for the equivalent of ``additional_pages``
    and ``next_page``.

    1. ``endpoint_template``: The parameteried URL for the API.
    2. ``additional_data_key``: The keypath to the value in the API response
       that determines whether there are additional pages to request.
    3. ``pagination_key``: The keypath to the value in the API response that
       contains the value that would be passed to the API to retrieve the next
       set of values.
    4. ``pagination_get_request_key``: The key in the ``endpoint_template``
       that will contain the value of the ``pagination_key``.

    For our simple example, the arguments would be

    1. ``endpoint_template: http://www.someapi.com/endpoint_name?resultpage={result_page}``
    2. ``additional_data_key: ["additional_pages"]``
    3. ``pagination_key: ["next_page"]``
    4. ``pagination_get_request_key: result_page``

    In addition to those mandatory arguments, you can also optionally specify
    an ``endpoint_dict``, which contains other values that will be substituted
    into the ``endpoint_template``. For example, these APIs often have an option
    that controls the number of results to provide in each response, like so:

    .. code-block:: none

        http://www.someapi.com/endpoint_name?results={num_results}?resultpage={result_page}

    For cases like this, the value of ``endpoint_dict`` is a dictionary mapping
    keys from the ``endpoint_template`` to their values. So if you wanted to
    have ten results per page, you would specify:

    .. code-block:: none

        endpoint_dict = {"num_results": 10}

    There can be any number of other parameters specified in the ``endpoint_dict``.

    If there are other keys in the ``endpoint_template`` that are not provided
    in the ``endpoint_dict``, then the node will try to find them in the current
    message that's being processed. For example, it is common to have some kind
    of security token that might be given in an environment variable. If the
    value of that environment variable has been provided by some upstream node
    and placed in the key ``token``, then it would be substituted into the
    URL, provded that the ``endpoint_template`` had a place for it, such as:

    .. code-block:: none

        http://www.someapi.com/endpoint_name?auth_token={token}?resultpage={result_page}

    """

    def __init__(
        self,
        endpoint_dict=None,
        json=True,
        pagination_get_request_key=None,
        endpoint_template=None,
        additional_data_key=None,
        pagination_key=None,
        pagination_template_key=None,
        default_offset_value="",
        **kwargs
    ):
        self.pagination_get_request_key = pagination_get_request_key
        self.additional_data_key = additional_data_key
        self.pagination_key = pagination_key
        self.endpoint_dict = endpoint_dict or {}
        self.endpoint_template = endpoint_template or ""
        self.default_offset_value = default_offset_value

        self.endpoint_template = self.endpoint_template.format_map(
            SafeMap(**self.endpoint_dict)
        )

        super(HttpGetRequestPaginator, self).__init__(**kwargs)

    def process_item(self):
        self.requestor = PaginatedHttpGetRequest(
            pagination_get_request_key=self.pagination_get_request_key,
            calling_node=self,
            endpoint_template=self.endpoint_template.format_map(
                SafeMap(**(self.message or {}))
            ),
            additional_data_key=self.additional_data_key,
            pagination_key=self.pagination_key,
            default_offset_value=self.default_offset_value,
        )

        for i in self.requestor.responses():
            logging.debug(
                "paginator GET request:"
                + str(self.name)
                + " "
                + str(self.requestor.endpoint_template)
            )
            logging.debug(
                "::".join([self.name, "events", str(len(i.get("events", "")))])
            )
            yield i
            if self.finished:
                break
